/* This file is part of Cloudy and is copyright (C)1978-2007 by Gary J. Ferland
 * For conditions of distribution and use see copyright notice in license.txt */
/*lines_helium put energetics, H, and He lines into line intensity stack */
/*TempInterp interpolates on a grid of values to produce predicted value at current Te.*/
#include "cddefines.h"
#include "taulines.h"
#include "dense.h"
#include "phycon.h"
#include "lines_service.h"
#include "iso.h"
#include "trace.h"
#include "lines.h"
#include "path.h"
#include "helike.h"
#include "thirdparty.h"
#include "physconst.h"
/*
#include "ipoint.h"
#include "rfield.h"
#include "radius.h"
#include "opacity.h"
*/
#define NUMTEMPS	22

typedef struct 
{
	/* index for upper and lower levels of line */
	long int ipHi;
	long int ipLo;

	char label[5];

} stdLines;

static void GetStandardHeLines(void);

static double TempInterp( double* TempArray , double* ValueArray, long NumElements );

static bool lgFirstRun = true;
static double CaABTemps[NUMTEMPS];
static long NumLines;
static double ***CaABIntensity;
static stdLines **CaABLines;

void lines_helium(void)
{
	long int i, j, 
	  ipHi, 
	  ipLo, 
	  nelem;
	char chLabel[5]="    ";
	double 
	  /*cb206, 
	  cb5016, 
	  cb5876 ,
	  em, 
	  rec,*/ 
	  sum,
	  Pop2_3S,
	  photons_3889_plus_7065 = 0.;

	DEBUG_ENTRY( "lines_helium()" );

	if( trace.lgTrace )
	{
		fprintf( ioQQQ, "   lines_helium called\n" );
	}

	/* HeI */
	i = StuffComment( "He-like iso-sequence" );
	linadd( 0., (float)i , "####", 'i');

#if	0
	/* He I 4471, */
	if( phycon.te <= 1e4 )
	{
		em = 2.42e-22/(phycon.te/phycon.te10);
	}
	else
	{
		em = 8.79e-22/phycon.te/phycon.te03/phycon.te01;
	}

	/* >>chng 97 jan 04, label was HeI, changed to Ca B for symmetry with others
	 * He I 4471 recombination only, fit to Brocklehurst '72 */
	rec = dense.eden*dense.xIonDense[ipHELIUM][1]*em;
	linadd( rec   , 4471 , "Ca B",'i');

	/* He I 5876 REC, simple fit to Brocklehurst */
	cb5876 = dense.eden*dense.xIonDense[ipHELIUM][1]*4.22e-21/(phycon.te*phycon.te10);
	linadd( cb5876 , 5876 , "Ca B" , 'i' );

	/* this is collisional enhancement of 5876 from 
	 * >>refer	He1	coll	KingdonJ., & Ferland, G.J. 1995, ApJ, 442, 714-725 */
	if( dense.eden < 1e8 )
	{
		double D = 1. + 3131.*100./phycon.sqrte / dense.eden;
		em = (6.78*0.52581*phycon.te07*sexp(3.776e4/phycon.te) +
			1.67*3.9811/(phycon.te10*phycon.te05)*sexp(4.545e4/phycon.te ) +
			0.60*22.9087/(phycon.te30*phycon.te04)*sexp(4.901e4/phycon.te) ) / D;
	}
	else
	{
		em = 0.;
	}
	/* simple He1 5876 with collisions */
	linadd( cb5876*(1.+em) , 5876 , "wCol" , 'i' );

	/* He I 6678 REC, simple fit to Brocklehurst */
	rec = dense.eden*dense.xIonDense[ipHELIUM][1]*1.32e-21/(phycon.te*phycon.te10*phycon.te01);
	linadd( rec , 6678 , "Ca B" ,'i' );

	em = dense.eden*dense.xIonDense[ipHELIUM][1];
	cb206 = 1.064e-22/phycon.te70/phycon.te10/phycon.te03/phycon.te03*
	  em;

	/* case B He I 2.06 micron - new or old wavelength scaling?  */
	linadd(cb206,20600.,"Ca B",'i');

	/* Case B He I 5016 */
	cb5016 = 5.43e-23/phycon.te70/phycon.te10*em;
	linadd( cb5016 , 5016 , "Ca B", 'i' );
#endif

#if	0
	{
		float DoublyIonizedWavelengths[18] = {
			18.96f, 19.2418f, 19.306f, 19.2831f,19.311f, 19.5747f, 19.2134f, 19.2186f, 19.439f,
			19.3217f,19.3126f,19.3217f,19.307f, 19.3179f,19.3269f, 19.5405f, 19.5498f, 19.5348f
		};
		double energyRyd[18] = {
			34.77502169,34.77502169,33.91347631,33.96960771,33.90125237,33.96960771,
			34.74903112,34.74903112,34.74903112,34.48322096,34.50539991,34.48322096,
			34.51915012,34.50539991,34.48322096,34.50539991,34.48322096,34.51915012};
		double stat_weight_hi[18] = {
			3.,3.,5.,3.,1.,3.,5.,5.,5.,
			3.,5.,3.,1.,5.,3.,5.,3.,1.
		};
		double stat_weight_lo[18] = {
			3.,1.,3.,3.,3.,1.,3.,5.,3.,
			1.,3.,3.,3.,5.,5.,3.,3.,3.
		};
		double branch_ratio[18] = {
			5.61E-05,1.00E+00,1.00E+00,1.00E+00,1.00E+00,5.60E-05,1.14E-06,4.36E-04,1.00E+00,
			1.11E-01,8.36E-02,8.36E-02,3.33E-01,2.50E-01,1.39E-01,1.02E-04,1.37E-05,2.77E-05
		};
		double sum_of_As[18] = {
			2.44014E+12,2.44014E+12,2.32013E+12,2.32013E+12,2.32013E+12,2.32013E+12,
			4.70206E+12,4.70206E+12,4.70206E+12,1.4002E+13,1.4002E+13,1.4002E+13,
			1.4002E+13,1.4002E+13,1.4002E+13,1.4002E+13,1.4002E+13,1.4002E+13
		};

		long nelem = ipOXYGEN;
		double total_dr_rate = 0.;

		for( i=0; i<18; i++ )
		{
			double ConBoltz, LTE_pop, factor1, ConvLTEPOP;
			double xInWrd;
			long ip;
			double temp;
			EmLine line1;

			EmLineZero( &line1 );
			
			factor1 = HION_LTE_POP*dense.AtomicWeight[nelem]/
				(dense.AtomicWeight[nelem]+ELECTRON_MASS/ATOMIC_MASS_UNIT);

			/* term in () is stat weight of electron * ion */
			ConvLTEPOP = pow(factor1,1.5)/(2.*iso.stat_ion[ipHE_LIKE])/phycon.te32;

			/* this Boltzmann factor is exp( +ioniz energy / Te ) */
			ConBoltz = dsexp(energyRyd[i]/phycon.te_ryd);

			if( ConBoltz >= SMALLDOUBLE )
			{
				/* >>chng 06 jul 18, the energy used to calculate ConBoltz above
				 * should be negative since this is above the continuum, but 
				 * to be safe we calculate ConBoltz with a positive energy above
				 * and multiply by it here instead of dividing.  */
				LTE_pop = stat_weight_hi[i] * ConBoltz * ConvLTEPOP;
			}
			
			temp = sum_of_As[i] * branch_ratio[i] * LTE_pop;
			total_dr_rate += temp;
			fprintf( ioQQQ, "%li\t%e\t%e\n", i, temp, total_dr_rate );

			line1.WLAng = DoublyIonizedWavelengths[i];
			line1.EnergyWN = (float)(RYD_INF*RYDLAM)/DoublyIonizedWavelengths[i];

			line1.phots = LTE_pop * sum_of_As[i] * branch_ratio[i] *
				dense.eden * dense.eden * dense.xIonDense[nelem][nelem];
			
			/* line1.phots /= 250.; */
			
			line1.xIntensity = line1.phots * EN1RYD * (RYDLAM/line1.WLAng);

			line1.ipCont = (int)ipLineEnergy(line1.EnergyWN * WAVNRYD , "O  7" , 
				iso.ipIsoLevNIonCon[ipHE_LIKE][ipOXYGEN][ipHe1s1S]);
			
			line1.ipFine = ipFineCont(line1.EnergyWN * WAVNRYD );

			line1.iRedisFun = ipCRDW;
			line1.nelem = ipOXYGEN + 1;
			line1.IonStg = 7;
			line1.Aul = (float)(sum_of_As[i] * branch_ratio[i]);
			line1.EnergyErg = (float)ERG1CM * line1.EnergyWN;
			line1.EnergyK = (float)T1CM * line1.EnergyWN;
			line1.gHi = (float)stat_weight_hi[i];
			line1.gLo = (float)stat_weight_lo[i];
			line1.gf = (float)GetGF(line1.Aul, line1.EnergyWN, line1.gHi);
			ASSERT(line1.gf > 0.);
			line1.FracInwd = 0.;
			
			PutLine(&line1);
			
			ip = line1.ipCont-1;

			/* inward fraction of line */
			xInWrd = line1.phots*line1.FracInwd;

			/* reflected part of line */
			rfield.reflin[ip] += (float)(xInWrd*radius.BeamInIn);

			/* inward beam that goes out since sphere set */
			/* in all this the ColOvTot term has been commented out,
				* since this is not meaningful for something like the hydrogen atom,
				* where most forms by recombination */
			rfield.outlin[ip] += (float)(xInWrd*radius.BeamInOut*opac.tmn[ip]);

			/* outward part of line */
			rfield.outlin[ip] += 
				(float)(line1.phots*(1. - line1.FracInwd)*
				radius.BeamOutOut*opac.tmn[ip]);
		}
		fprintf( ioQQQ, "dr rate\t%e\t%e\t%e\n\n", 
			total_dr_rate, dense.eden, dense.xIonDense[nelem][nelem] );
	}
#endif

	/* read in Case A and B lines from data file	*/
	if( lgFirstRun )
	{
		GetStandardHeLines();
		lgFirstRun = false;
	}

	/* this is the main printout, where line intensities are entered into the stack */
	for( nelem=1; nelem < LIMELM; nelem++ )
	{
		if( dense.lgElmtOn[nelem] )
		{
			if( nelem == ipHELIUM )
			{
				double *qTotEff;

				/* >>chng 06 aug 17, all of these from _max to _local */
				/* >>chng 06 dec 21, mistake - change back to _max */
				qTotEff = (double*)MALLOC(sizeof(double)*(unsigned)(iso.numLevels_max[ipHE_LIKE][nelem]) );

				qTotEff[0] = 0.;
				qTotEff[1] = 0.;
					
				for( i=ipHe2s3S+1; i<iso.numLevels_max[ipHE_LIKE][nelem]-iso.nCollapsed_max[ipHE_LIKE][nelem]; i++ )
				{
					qTotEff[i] = 0.;
					for( j = i; j<iso.numLevels_max[ipHE_LIKE][nelem]-iso.nCollapsed_max[ipHE_LIKE][nelem]; j++ )
					{
						/*if( iso.quant_desig[ipHE_LIKE][nelem][i].s == 1 )
						{*/
							qTotEff[i] += 
								EmisLines[ipHE_LIKE][nelem][j][ipHe2s3S].ColUL*dense.EdenHCorr*
								iso.Boltzmann[ipHE_LIKE][nelem][j][ipHe2s3S] *
								(double)EmisLines[ipHE_LIKE][nelem][j][ipHe2s3S].gHi / 
								(double)EmisLines[ipHE_LIKE][nelem][j][ipHe2s3S].gLo*
								helike.CascadeProb[nelem][j][i];
						/*}*/
					}
				}

				/* print simple 2^3S population and the one cloudy got	*/
				/* chng 05 dec 14, RadRec_caseB now contains only recombinations into triplets,
				 * so no need for the 0.75, which is inaccurate anyway */
				Pop2_3S = dense.eden*iso.RadRec_caseB[ipHE_LIKE][nelem]/
					( EmisLines[ipHE_LIKE][nelem][ipHe2s3S][ipHe1s1S].Aul+ dense.eden*helike.qTot2TripS[nelem]);

				for( i=0; i< NumLines; i++ )
				{
					ipHi = CaABLines[nelem][i].ipHi;
					ipLo = CaABLines[nelem][i].ipLo;

					/* >>chng 06 aug 17, from _max to _local */
					/* >>chng 06 dec 21, mistake - change back to _max */
					if( ipHi <= iso.n_HighestResolved_max[ipHE_LIKE][nelem]*(iso.n_HighestResolved_max[ipHE_LIKE][nelem]+1))
					{
						double intens = TempInterp( CaABTemps , CaABIntensity[nelem][i], NUMTEMPS )*
							dense.xIonDense[nelem][nelem]*dense.eden;

						linadd( intens,
							EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].WLAng,
							CaABLines[nelem][i].label,
							'i');

						if( nMatch("Ca B",CaABLines[nelem][i].label) )
						{
							/* chng 05 dec 14, branching ratio was missing here! 
							 * not a big effect because lines with biggest collision
							 * enhancements tend to be dominant decay route from upper level. */
							linadd( intens + 
								Pop2_3S*qTotEff[ipHi]*dense.xIonDense[nelem][nelem]*
								helike.BranchRatio[nelem][ipHi][ipLo]*
								EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].EnergyErg,
								EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].WLAng,
								"+Col",
								'i');
						}
					}
					/* Make sure to at least do 4471 */
					else if( ipLo==ipHe2p3P1 && ipHi==ipHe4d3D && nMatch("Ca B",CaABLines[nelem][i].label) )
					{
						double intens = TempInterp( CaABTemps , CaABIntensity[nelem][i], NUMTEMPS )*
							dense.xIonDense[nelem][nelem]*dense.eden;

						linadd( intens, 4471, CaABLines[nelem][i].label, 'i');
					}

				}
				free( qTotEff );
			}

			/* NB NB - low and high must be in this order so that all balmer, paschen,
			 * etc series line up correctly in final printout */
			/* >>chng 01 jun 13, bring 23P lines back together */
			/* two photon is special, not a line and no ipCont array index, add here */

			EmisLines[ipHE_LIKE][nelem][ipHe2s1S][ipHe1s1S].xIntensity = 
				(double)EmisLines[ipHE_LIKE][nelem][ipHe2s1S][ipHe1s1S].Aul*
				(double)EmisLines[ipHE_LIKE][nelem][ipHe2s1S][ipHe1s1S].PopHi*
				(double)EmisLines[ipHE_LIKE][nelem][ipHe2s1S][ipHe1s1S].Pesc*
				(double)dense.xIonDense[nelem][nelem]*
				(double)EmisLines[ipHE_LIKE][nelem][ipHe2s1S][ipHe1s1S].EnergyErg;
			if( LineSave.ipass == 0 )
			{
				/* chIonLbl is function that generates a null terminated 4 char string, of form "C  2" 
				 * the result, chLable, is only used when ipass == 0, can be undefined otherwise */
				/* total two photon emission */
				chIonLbl(chLabel, &EmisLines[ipHE_LIKE][nelem][ipHe2s1S][ipHe1s1S]);
			}
			/* two photon continuum */
			linadd( EmisLines[ipHE_LIKE][nelem][ipHe2s1S][ipHe1s1S].xIntensity , 0,chLabel,'r');

			/* induced two photon emission */
			linadd(
				iso.Pop2Ion[ipHE_LIKE][nelem][ipHe2s1S]*
				dense.xIonDense[nelem][nelem]*
				iso.TwoNu_induc_dn[ipHE_LIKE][nelem]*
				EmisLines[ipHE_LIKE][nelem][ipHe2s1S][ipHe1s1S].EnergyErg,
				22, chLabel ,'i');

			/* here we will create an entry for the three lines 
			 * coming from 2 3P to 1 1S - the entry called TOTL will
			 * appear before the lines of the multiplet */
			sum = 0.;
			for( i=ipHe2p3P0; i <= ipHe2p3P2; i++ )
			{
				sum += 
					(double)EmisLines[ipHE_LIKE][nelem][i][ipHe1s1S].Aul*
					(double)EmisLines[ipHE_LIKE][nelem][i][ipHe1s1S].PopHi*
					(double)EmisLines[ipHE_LIKE][nelem][i][ipHe1s1S].Pesc*
					(double)dense.xIonDense[nelem][nelem]*
					(double)EmisLines[ipHE_LIKE][nelem][i][ipHe1s1S].EnergyErg;
			}

			/* total emission in He-like forbidden lines from 2p3P to ground */
			linadd(sum,EmisLines[ipHE_LIKE][nelem][ipHe2p3P1][ipHe1s1S].WLAng,"TOTL",'i' );
			
			/* now do real permitted lines */
			for( ipLo=0; ipLo < ipHe2p3P0; ipLo++ )
			{
				for( ipHi=ipLo+1; ipHi < iso.numPrintLevels[ipHE_LIKE][nelem]; ipHi++ )
				{
					/* >>chng 01 may 30, do not add fake he-like lines (majority) to line stack */
					/* >>chng 01 dec 11, use variable for smallest A */
					if( EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].ipCont < 1 ) 
						continue;

					/* recombine fine-structure lines since the energies are
					 * not resolved anyway.	*/
					if( helike.lgFSM && ( abs(iso.quant_desig[ipHE_LIKE][nelem][ipHi].l -
						iso.quant_desig[ipHE_LIKE][nelem][ipLo].l)==1 )
						&& (iso.quant_desig[ipHE_LIKE][nelem][ipLo].l>1) 
						&& (iso.quant_desig[ipHE_LIKE][nelem][ipHi].l>1) 
						&& ( iso.quant_desig[ipHE_LIKE][nelem][ipHi].n ==
						iso.quant_desig[ipHE_LIKE][nelem][ipLo].n ) )
					{
						if( (iso.quant_desig[ipHE_LIKE][nelem][ipHi].s==0) 
							&& (iso.quant_desig[ipHE_LIKE][nelem][ipLo].s==0) )
						{
							continue;
						}
						else if( (iso.quant_desig[ipHE_LIKE][nelem][ipHi].s==1) 
							&& (iso.quant_desig[ipHE_LIKE][nelem][ipLo].s==1) )
						{

							/* singlet to singlet*/
							EmisLines[ipHE_LIKE][nelem][ipHi+1][ipLo+1].phots = 
								( (double)EmisLines[ipHE_LIKE][nelem][ipHi][ipLo+1].Aul*
								(double)EmisLines[ipHE_LIKE][nelem][ipHi][ipLo+1].PopHi*
								(double)EmisLines[ipHE_LIKE][nelem][ipHi][ipLo+1].Pesc +
								(double)EmisLines[ipHE_LIKE][nelem][ipHi+1][ipLo+1].Aul*
								(double)EmisLines[ipHE_LIKE][nelem][ipHi+1][ipLo+1].PopHi*
								(double)EmisLines[ipHE_LIKE][nelem][ipHi+1][ipLo+1].Pesc )*
								(double)dense.xIonDense[nelem][nelem];
							
							EmisLines[ipHE_LIKE][nelem][ipHi+1][ipLo+1].xIntensity = 
								EmisLines[ipHE_LIKE][nelem][ipHi+1][ipLo+1].phots *
								(double)EmisLines[ipHE_LIKE][nelem][ipHi+1][ipLo+1].EnergyErg;

							PutLine(&EmisLines[ipHE_LIKE][nelem][ipHi+1][ipLo+1]);
						
							/* triplet to triplet */
							EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].phots = 
								( (double)EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].Aul*
								(double)EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].PopHi*
								(double)EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].Pesc +
								(double)EmisLines[ipHE_LIKE][nelem][ipHi+1][ipLo].Aul*
								(double)EmisLines[ipHE_LIKE][nelem][ipHi+1][ipLo].PopHi*
								(double)EmisLines[ipHE_LIKE][nelem][ipHi+1][ipLo].Pesc )*
								(double)dense.xIonDense[nelem][nelem];
							
							EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].xIntensity = 
								EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].phots *
								(double)EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].EnergyErg;

							PutLine(&EmisLines[ipHE_LIKE][nelem][ipHi][ipLo]);
						}
					}
					
					else if( ipLo==ipHe2s3S && ipHi == ipHe2p3P0 )
					{
						/* here we will create an entry for the three lines 
						 * coming from 2 3P to 2 3S - the entry called TOTL will
						 * appear before the lines of the multiplet 
						 * for He I this is 10830 */

						float av_wl = 0.;
						sum = 0.;
						for( i=ipHe2p3P0; i <= ipHe2p3P2; i++ )
						{
							sum += 
								(double)EmisLines[ipHE_LIKE][nelem][i][ipLo].Aul*
								(double)EmisLines[ipHE_LIKE][nelem][i][ipLo].PopHi*
								(double)EmisLines[ipHE_LIKE][nelem][i][ipLo].Pesc*
								(double)dense.xIonDense[nelem][nelem]*
								(double)EmisLines[ipHE_LIKE][nelem][i][ipLo].EnergyErg;
							av_wl += EmisLines[ipHE_LIKE][nelem][i][ipLo].WLAng;
						}
						av_wl /= 3.;
#						if 0
						{
#						include "elementnames.h"
#						include "prt.h"
						fprintf(ioQQQ,"DEBUG 2P - 2S multiplet wl %s ",
							elementnames.chElementSym[nelem] );
						prt_wl( ioQQQ , av_wl );
						fprintf(ioQQQ,"\n" );
						}
#						endif

						/* total emission in He-like lines 
						 * >>chng 06 aug 18, use average of three line wavelengths */
						linadd(sum,av_wl,"TOTL",'i' );
#						ifdef PRT_HELIKE
#						endif
						
						/*>>chng 05 sep 8, added the following so that the component
						 * from ipHe2p3P0 is printed, in addition to the total. */
						
						/* find number of photons escaping cloud */
						EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].phots = 
							EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].Aul*
							EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].PopHi*
							EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].Pesc*
							dense.xIonDense[nelem][nelem];

						/* now find line intensity */
						EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].xIntensity = 
							(double)EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].Aul*
							(double)EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].PopHi*
							(double)EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].Pesc*
							(double)dense.xIonDense[nelem][nelem]*
							(double)EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].EnergyErg;
						
						if( helike.lgRandErrGen )
						{
							EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].phots *=
								helike.ErrorFactor[nelem][ipHi][ipLo][IPRAD];
							EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].xIntensity *= 
								helike.ErrorFactor[nelem][ipHi][ipLo][IPRAD];
						}
						PutLine(&EmisLines[ipHE_LIKE][nelem][ipHi][ipLo]);
					}

					else
					{

						/* find number of photons escaping cloud */
						EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].phots = 
							EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].Aul*
							EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].PopHi*
							EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].Pesc*
							dense.xIonDense[nelem][nelem];

						/* now find line intensity */
						/* >>chng 01 jan 15, put cast double to force double evaluation */
						EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].xIntensity = 
							(double)EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].Aul*
							(double)EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].PopHi*
							(double)EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].Pesc*
							(double)dense.xIonDense[nelem][nelem]*
							(double)EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].EnergyErg;

						if( helike.lgRandErrGen )
						{
							EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].phots *=
								helike.ErrorFactor[nelem][ipHi][ipLo][IPRAD];
							EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].xIntensity *= 
								helike.ErrorFactor[nelem][ipHi][ipLo][IPRAD];
						}

						/* this will enter .xIntensity into the line stack
						fprintf(ioQQQ,"1 loop %li %li %.1f\n", ipLo,ipHi, 
							EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].WLAng ); */
						PutLine(&EmisLines[ipHE_LIKE][nelem][ipHi][ipLo]);
						{
							/* option to print particulars of some line when called
							 * a prettier print statement is near where chSpin is defined below*/
							/*@-redef@*/
						enum {DEBUG_LOC=false};
							/*@+redef@*/
						if( DEBUG_LOC )
							{
								if( nelem==1 && ipLo==0 && ipHi==1 )
								{
									fprintf(ioQQQ,"he1 626 %.2e %.2e \n", 
										EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].TauIn,
										EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].TauTot
										);
								}
							}
						}
					}
				}
			}

			/* this sum will bring together the three lines going to J levels within 23P */
			for( ipHi=ipHe2p3P2+1; ipHi < iso.numPrintLevels[ipHE_LIKE][nelem]; ipHi++ )
			{
				double sumcool , sumheat ,
					save , savecool , saveheat;

				sum = 0;
				sumcool = 0.;
				sumheat = 0.;
				for( ipLo=ipHe2p3P0; ipLo <= ipHe2p3P2; ++ipLo )
				{
					/* find number of photons escaping cloud */
					EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].phots = 
						EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].Aul*
						EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].PopHi*
						EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].Pesc*
						dense.xIonDense[nelem][nelem];

					/* now find line intensity */
					/* >>chng 01 jan 15, put cast double to force double evaluation */
					EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].xIntensity = 
						(double)EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].Aul*
						(double)EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].PopHi*
						(double)EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].Pesc*
						(double)dense.xIonDense[nelem][nelem]*
						(double)EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].EnergyErg;

					if( helike.lgRandErrGen )
					{
						EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].phots *=
							helike.ErrorFactor[nelem][ipHi][ipLo][IPRAD];
						EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].xIntensity *= 
							helike.ErrorFactor[nelem][ipHi][ipLo][IPRAD];
					}

					sumcool += EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].cool;
					sumheat += EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].heat;
					sum += EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].xIntensity;
				}

				/* >>chng 01 dec 11, use variable for smallest A */
				/* >>chng 02 feb 10, replaced check on A with ipCont */
				if( EmisLines[ipHE_LIKE][nelem][ipHi][ipHe2p3P2].ipCont < 1 ) 
					continue;
				/* this will enter .xIntensity into the line stack */
				save = EmisLines[ipHE_LIKE][nelem][ipHi][ipHe2p3P2].xIntensity;
				savecool = EmisLines[ipHE_LIKE][nelem][ipHi][ipHe2p3P2].cool;
				saveheat = EmisLines[ipHE_LIKE][nelem][ipHi][ipHe2p3P2].heat;

				EmisLines[ipHE_LIKE][nelem][ipHi][ipHe2p3P2].xIntensity = sum;
				EmisLines[ipHE_LIKE][nelem][ipHi][ipHe2p3P2].cool = sumcool;
				EmisLines[ipHE_LIKE][nelem][ipHi][ipHe2p3P2].heat = sumheat;

				/*fprintf(ioQQQ,"2 loop %li %li %.1f\n", ipHe2p3P2,ipHi, 
					EmisLines[ipHE_LIKE][nelem][ipHi][ipHe2p3P2].WLAng );*/
				PutLine(&EmisLines[ipHE_LIKE][nelem][ipHi][ipHe2p3P2]);

				EmisLines[ipHE_LIKE][nelem][ipHi][ipHe2p3P2].xIntensity = save;
				EmisLines[ipHE_LIKE][nelem][ipHi][ipHe2p3P2].cool = savecool;
				EmisLines[ipHE_LIKE][nelem][ipHi][ipHe2p3P2].heat = saveheat;
			}
			for( ipLo=ipHe2p3P2+1; ipLo < iso.numPrintLevels[ipHE_LIKE][nelem]-1; ipLo++ )
			{
				for( ipHi=ipLo+1; ipHi < iso.numPrintLevels[ipHE_LIKE][nelem]; ipHi++ )
				{
					/* >>chng 01 may 30, do not add fake he-like lines (majority) to line stack */
					/* >>chng 01 dec 11, use variable for smallest A */
					/* >>chng 02 feb 12, use ipCont to find bogus lines */
					/*if( EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].Aul <= iso.SmallA )*/
					if( EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].ipCont < 1 ) 
						continue;

					/* find number of photons escaping cloud */
					EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].phots = 
						EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].Aul*
						EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].PopHi*
						EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].Pesc*
						dense.xIonDense[nelem][nelem];

					/* now find line intensity */
					/* >>chng 01 jan 15, put cast double to force double evaluation */
					EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].xIntensity = 
						(double)EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].Aul*
						(double)EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].PopHi*
						(double)EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].Pesc*
						(double)dense.xIonDense[nelem][nelem]*
						(double)EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].EnergyErg;

					if( helike.lgRandErrGen )
					{
						EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].phots *=
							helike.ErrorFactor[nelem][ipHi][ipLo][IPRAD];
						EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].xIntensity *= 
							helike.ErrorFactor[nelem][ipHi][ipLo][IPRAD];
					}

					/* this will enter .xIntensity into the line stack 
					fprintf(ioQQQ,"2 loop %li %li %.1f\n", ipLo,ipHi, 
						EmisLines[ipHE_LIKE][nelem][ipHi][ipLo].WLAng );*/
					PutLine(&EmisLines[ipHE_LIKE][nelem][ipHi][ipLo]);
				}
			}
		}
	}

	if( iso.n_HighestResolved_max[ipHE_LIKE][ipHELIUM] >= 4 &&
		iso.n_HighestResolved_max[ipH_LIKE][ipHYDROGEN] >=8 )
	{
		/* For info only, add the total photon flux in 3889 and 7065,
		* and 3188, 4713, and 5876. */
		photons_3889_plus_7065 =
			/* to 2p3P2 */
			EmisLines[ipHE_LIKE][ipHELIUM][ipHe3s3S][ipHe2p3P2].xIntensity/
			EmisLines[ipHE_LIKE][ipHELIUM][ipHe3s3S][ipHe2p3P2].EnergyErg +
			EmisLines[ipHE_LIKE][ipHELIUM][ipHe3d3D][ipHe2p3P2].xIntensity/
			EmisLines[ipHE_LIKE][ipHELIUM][ipHe3d3D][ipHe2p3P2].EnergyErg +
			EmisLines[ipHE_LIKE][ipHELIUM][ipHe4s3S][ipHe2p3P2].xIntensity/
			EmisLines[ipHE_LIKE][ipHELIUM][ipHe4s3S][ipHe2p3P2].EnergyErg +
			/* to 2p3P1 */
			EmisLines[ipHE_LIKE][ipHELIUM][ipHe3s3S][ipHe2p3P1].xIntensity/
			EmisLines[ipHE_LIKE][ipHELIUM][ipHe3s3S][ipHe2p3P1].EnergyErg +
			EmisLines[ipHE_LIKE][ipHELIUM][ipHe3d3D][ipHe2p3P1].xIntensity/
			EmisLines[ipHE_LIKE][ipHELIUM][ipHe3d3D][ipHe2p3P1].EnergyErg +
			EmisLines[ipHE_LIKE][ipHELIUM][ipHe4s3S][ipHe2p3P1].xIntensity/
			EmisLines[ipHE_LIKE][ipHELIUM][ipHe4s3S][ipHe2p3P1].EnergyErg +
			/* to 2p3P0 */
			EmisLines[ipHE_LIKE][ipHELIUM][ipHe3s3S][ipHe2p3P0].xIntensity/
			EmisLines[ipHE_LIKE][ipHELIUM][ipHe3s3S][ipHe2p3P0].EnergyErg +
			EmisLines[ipHE_LIKE][ipHELIUM][ipHe3d3D][ipHe2p3P0].xIntensity/
			EmisLines[ipHE_LIKE][ipHELIUM][ipHe3d3D][ipHe2p3P0].EnergyErg +
			EmisLines[ipHE_LIKE][ipHELIUM][ipHe4s3S][ipHe2p3P0].xIntensity/
			EmisLines[ipHE_LIKE][ipHELIUM][ipHe4s3S][ipHe2p3P0].EnergyErg +
			/* to 2s3S */
			EmisLines[ipHE_LIKE][ipHELIUM][ipHe3p3P][ipHe2s3S].xIntensity/
			EmisLines[ipHE_LIKE][ipHELIUM][ipHe3p3P][ipHe2s3S].EnergyErg +
			EmisLines[ipHE_LIKE][ipHELIUM][ipHe4p3P][ipHe2s3S].xIntensity/
			EmisLines[ipHE_LIKE][ipHELIUM][ipHe4p3P][ipHe2s3S].EnergyErg;

		/* Add in photon flux of H8 3889 */
		photons_3889_plus_7065 += 
			EmisLines[ipH_LIKE][ipHYDROGEN][8][1].xIntensity/
			EmisLines[ipH_LIKE][ipHYDROGEN][8][1].EnergyErg;
		/* now multiply by ergs of normalization line, so that relative flux of
		* this line will be ratio of photon fluxes. */
		photons_3889_plus_7065 *= (ERG1CM*1.e8)/LineSave.WavLNorm;
		linadd( photons_3889_plus_7065, 3889., "Pho+", 'i');
	}

	/* ====================================================
	 * end helium
	 * ====================================================*/

#if	0
	/* Need to free these things, but only one time at the end. */
	/* Free some malloc'd arrays */
	for( nelem=ipHELIUM; nelem<LIMELM; ++nelem )
	{
		/** \todo	2	- this structure is currently only used for helium itself...
		 * stuff numbers in for other elements, or drop the [nelem] dimension off
		 * of CaABLines	*/
		if( nelem != ipHELIUM )
			continue;

		/* only grab core for elements that are turned on */
		if( nelem == ipHELIUM || dense.lgElmtOn[nelem])
		{
			for( j = 0; j < NumLines; ++j )
			{
				free( CaABIntensity[nelem][j] );
			}
			free( CaABIntensity[nelem] );
			free( CaABLines[nelem] );
		}
	}
	free( CaABIntensity );
	free( CaABLines );
#endif

	if( trace.lgTrace )
	{
		fprintf( ioQQQ, "   lines_helium returns\n" );
	}

	DEBUG_EXIT( "lines_helium()" );
	return;
}


static void GetStandardHeLines(void)
{
	FILE *ioDATA;
	bool lgEOL, lgHIT;
	long i, i1, i2, j, nelem;

#	define chLine_LENGTH 1000
	char chLine[chLine_LENGTH] , 
		/* this must be longer than chDataPath, set in path.h */
		chFilename[FILENAME_PATH_LENGTH_2];

	/* check on path if file not here and path set */
	/* path was parsed in getset */
	if( lgDataPathSet == true )
	{
		/*path set, so look only there */
		strcpy( chFilename , chDataPath );
		strcat( chFilename , "he1_case_ab.dat" );
	}
	else
	{
		/* path not set, check local space only */
		strcpy( chFilename , "he1_case_ab.dat" );
	}

	if( trace.lgTrace )
		fprintf( ioQQQ," lines_helium opening he1_case_ab.dat:");

	if( ( ioDATA = fopen( chFilename , "r" ) ) == NULL )
	{
		fprintf( ioQQQ, " lines_helium could not open he1_case_ab.dat\n" );
		if( lgDataPathSet == true )
			fprintf( ioQQQ, " even tried path\n" );

		if( lgDataPathSet == true )
		{
			fprintf( ioQQQ, " lines_helium could not open he1_case_ab.dat\n");
			fprintf( ioQQQ, " path is ==%s==\n",chDataPath );
			fprintf( ioQQQ, " final path is ==%s==\n",chFilename );
		}

		puts( "[Stop in lines_helium]" );
		cdEXIT(EXIT_FAILURE);
	}

	/* check that magic number is ok */
	if( fgets( chLine , (int)sizeof(chLine) , ioDATA ) == NULL )
	{
		fprintf( ioQQQ, " lines_helium could not read first line of he1_case_ab.dat.\n");
		puts( "[Stop in lines_helium]" );
		cdEXIT(EXIT_FAILURE);
	}
	i = 1;
	/* first number is magic number, second is number of lines in file	*/
	i1 = (long)FFmtRead(chLine,&i,INPUT_LINE_LENGTH,&lgEOL);
	i2 = (long)FFmtRead(chLine,&i,INPUT_LINE_LENGTH,&lgEOL);
	NumLines = i2;

	/* the following is to check the numbers that appear at the start of he1_case_ab.dat */
	if( i1 !=CASEABMAGIC )
	{
		fprintf( ioQQQ, 
			" lines_helium: the version of he1_case_ab.dat is not the current version.\n" );
		fprintf( ioQQQ, 
			" lines_helium: I expected to find the number %i and got %li instead.\n" ,
			CASEABMAGIC, i1 );
		fprintf( ioQQQ, "Here is the line image:\n==%s==\n", chLine );
		puts( "[Stop in lines_helium]" );
		cdEXIT(EXIT_FAILURE);
	}

	/* get the array of temperatures */
	lgHIT = false;
	while( fgets( chLine , (int)sizeof(chLine) , ioDATA ) != NULL )
	{
		/* only look at lines without '#' in first col */
		if( chLine[0] != '#')
		{
			lgHIT = true;
			j = 0;
			lgEOL = false;
			i = 1;
			while( !lgEOL && j < NUMTEMPS)
			{
				CaABTemps[j] = FFmtRead(chLine,&i,INPUT_LINE_LENGTH,&lgEOL);
				++j;
			}
			break;
		}
	}

	if( !lgHIT )
	{
		fprintf( ioQQQ, " lines_helium could not find line of temperatures in he1_case_ab.dat.\n");
		puts( "[Stop in lines_helium]" );
		cdEXIT(EXIT_FAILURE);
	}

	/* create space for array of values, if not already done */
	CaABIntensity = (double ***)MALLOC(sizeof(double **)*(unsigned)LIMELM );
	/* create space for array of values, if not already done */
	CaABLines = (stdLines **)MALLOC(sizeof(stdLines *)*(unsigned)LIMELM );

	for( nelem=ipHELIUM; nelem<LIMELM; ++nelem )
	{
		/** \todo	2	- this structure is currently only used for helium itself...
		 * stuff numbers in for other elements, or drop the [nelem] dimension off
		 * of CaABLines	*/
		if( nelem != ipHELIUM )
			continue;

		/* only grab core for elements that are turned on */
		if( nelem == ipHELIUM || dense.lgElmtOn[nelem])
		{
			/* create space for array of values, if not already done */
			CaABIntensity[nelem] = (double **)MALLOC(sizeof(double *)*(unsigned)(i2) );
			CaABLines[nelem] = (stdLines *)MALLOC(sizeof(stdLines )*(unsigned)(i2) );

			/* avoid allocating 0 bytes, some OS return NULL pointer, PvH 
			CaABIntensity[nelem][0] = NULL;*/
			for( j = 0; j < i2; ++j )
			{
				CaABIntensity[nelem][j] = (double *)MALLOC(sizeof(double)*(unsigned)NUMTEMPS );

				CaABLines[nelem][j].ipHi = -1;
				CaABLines[nelem][j].ipLo = -1;
				strcpy( CaABLines[nelem][j].label , "    " );

				for( i=0; i<NUMTEMPS; ++i )
				{
					CaABIntensity[nelem][j][i] = 0.;
				}
			}
		}
	}

	/* now read in the case A and B line data */
	lgHIT = false;
	nelem = ipHELIUM;
	while( fgets( chLine , (int)sizeof(chLine) , ioDATA ) != NULL )
	{
		static long line = 0;
		char *chTemp;
				
		/* only look at lines without '#' in first col */
		if( (chLine[0] == ' ') || (chLine[0]=='\n') )
			break;
		if( chLine[0] != '#')
		{
			lgHIT = true;

			/* get lower and upper level index */
			j = 1;
			/* the first number is the wavelength, which is not used
			 * for anything, but must skip over it. */
			i1 = (long)FFmtRead(chLine,&j,INPUT_LINE_LENGTH,&lgEOL);
			CaABLines[nelem][line].ipLo = (long)FFmtRead(chLine,&j,INPUT_LINE_LENGTH,&lgEOL);
			CaABLines[nelem][line].ipHi = (long)FFmtRead(chLine,&j,INPUT_LINE_LENGTH,&lgEOL);
			
			ASSERT( CaABLines[nelem][line].ipLo < CaABLines[nelem][line].ipHi );
			/*if( CaABLines[nelem][line].ipHi >= iso.numLevels_max[ipHE_LIKE][nelem] )
				continue;*/

			chTemp = chLine;
			/* skip over 4 tabs to start of data */
			for( i=0; i<3; ++i )
			{
				if( (chTemp = strchr( chTemp, '\t' )) == NULL )
				{
					fprintf( ioQQQ, " lines_helium no init case A and B\n" );
					puts( "[Stop in lines_helium]" );
					cdEXIT(EXIT_FAILURE);
				}
				++chTemp;
			}

			strncpy( CaABLines[nelem][line].label, chTemp , 4 );
			CaABLines[nelem][line].label[4] = 0;
				
			for( i=0; i<NUMTEMPS; ++i )
			{
				float b;
				if( (chTemp = strchr( chTemp, '\t' )) == NULL )
				{
					fprintf( ioQQQ, " lines_helium could not scan case A and B lines, current indices: %li %li\n",
						CaABLines[nelem][line].ipHi,
						CaABLines[nelem][line].ipLo );
					puts( "[Stop in lines_helium]" );
					cdEXIT(EXIT_FAILURE);
				}
				++chTemp;
				sscanf( chTemp , "%e" , &b );
				CaABIntensity[nelem][line][i] = pow(10.,(double)b);
			}
			line++;
		}
	}

	/* close the data file */
	fclose( ioDATA );

	return;
}

/** \todo	there is a virtually identical routine in helike_recom.cpp -> combine */
static double TempInterp( double* TempArray , double* ValueArray, long NumElements )
{
	long int ipTe=-1;
	double rate = 0.;
	long i0;

	DEBUG_ENTRY( "TempInterp()" );

	ASSERT( fabs( 1. - (double)phycon.alogte/log10(phycon.te) ) < 0.0001 );

	/* te totally unknown */
	if( phycon.alogte <= TempArray[0] )
	{
		return ValueArray[0];
	}
	else if( phycon.alogte >= TempArray[NumElements-1] )
	{
		return ValueArray[NumElements-1];
	}

	/* now search for temperature */
	ipTe = hunt_bisect( TempArray, NumElements, phycon.alogte );			

	ASSERT( (ipTe >=0) && (ipTe < NumElements-1)  );
			
#if	0
	rate = ((double)phycon.alogte - TempArray[ipTe]) / (TempArray[ipTe+1] - TempArray[ipTe]) *
		(ValueArray[ipTe+1] - ValueArray[ipTe]) + ValueArray[ipTe];
#endif
	
	/* Do a four-point interpolation */
	const int ORDER = 3; /* order of the fitting polynomial */
	i0 = max(min(ipTe-ORDER/2,NumElements-ORDER-1),0);
	rate = lagrange( &TempArray[i0], &ValueArray[i0], ORDER+1, phycon.alogte );

	DEBUG_EXIT( "TempInterp()" );

	return rate;
}
